#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


from argparse import ArgumentParser

from hans.plottools import DatasetSelector


def get_parser():

    parser = ArgumentParser()
    parser.add_argument('-p', dest="path", default=".", help="path (default: data)")

    return parser


if __name__ == "__main__":

    parser = get_parser()
    args = parser.parse_args()

    files = DatasetSelector(args.path)

    for filename, data in files._ds.items():

        print(filename + ": \n" + 40 * "-")
        for name in data.ncattrs():
            print("{:20s}: {:>}".format(name, getattr(data, name)))
        print(40 * "-")



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from argparse import ArgumentParser
import matplotlib.pyplot as plt
import matplotlib.ticker as tk

from hans.plottools import DatasetSelector


def get_parser():
    parser = ArgumentParser()
    parser.add_argument('-p', dest="path", default="data", help="path (default: data)")
    parser.add_argument('-v', dest="key", default=None, choices=[None, "rho", "p", "jx", "jy"], help="variable (default: None)")
    parser.add_argument('-d', dest="dir", default="x", choices=["x", "y"], help="cutting direction (default: x)")
    parser.add_argument('-f', dest="freq", type=int, default=1, help="plot frequency (default: 1)")

    return parser


if __name__ == "__main__":

    ylabels = {"rho": r"Density $\rho$",
               "p": r"Pressure $p$",
               "jx": r"Momentum density $j_x$",
               "jy": r"Momentum denisty $j_y$"}

    cmap = plt.cm.coolwarm

    parser = get_parser()
    args = parser.parse_args()

    files = DatasetSelector(args.path, mode="single")
    fn, = files.get_filenames()
    time, xdata, ydata = files.get_centerlines(key=args.key, freq=args.freq, dir=args.dir)[0]
    print("Plotting ", fn)

    if args.key is None:
        fig, ax = plt.subplots(2, 2, sharex=True, figsize=(12.8, 9.6))

        ax[1, 0].set_xlabel(rf"Distance ${args.dir}$")
        ax[1, 1].set_xlabel(rf"Distance ${args.dir}$")
        maxT = time[-1]
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=0, vmax=maxT))

        for key, axis in zip(ydata.keys(), ax.flat):
            for i, t in enumerate(time):
                axis.plot(xdata, ydata[key][i], color=cmap(t / maxT))
            axis.set_ylabel(ylabels[key])

        fmt = tk.ScalarFormatter(useMathText=True)
        fmt.set_powerlimits((0, 0))
        fig.colorbar(sm, ax=ax.ravel().tolist(), format=fmt, label='time $t$', extend='max', orientation="horizontal", aspect=50, pad=0.1)

    else:
        fig, ax = plt.subplots(1, figsize=(6.4, 4.8), tight_layout=True)

        maxT = time[-1]
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=0, vmax=maxT))

        for i, t in enumerate(time):
            ax.plot(xdata, ydata[i], color=cmap(t/maxT))
            ax.set_ylabel(ylabels[args.key])
            ax.set_xlabel(rf"Distance ${args.dir}$")

        fmt = tk.ScalarFormatter(useMathText=True)
        fmt.set_powerlimits((0, 0))
        fig.colorbar(sm, ax=ax, format=fmt, label='time $t$', extend='max', orientation="horizontal", aspect=50, pad=0.1)

    plt.show()



#
# Copyright 2019, 2022 Hannes Holey
#           2019 Andrea Codrignani
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import os
import signal
from netCDF4 import Dataset
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from datetime import datetime
from pkg_resources import get_distribution
import numpy as np
import shutil

from hans.tools import abort
from hans.material import Material
from hans.plottools import adaptiveLimits
from hans.integrate import ConservedField


class Problem:

    def __init__(self, options, disc, bc, geometry, numerics, material, surface, ic, roughness):
        """
        Collects all information about a single problem
        and contains the methods to run a simulation, based on the problem defintiion."

        Parameters
        ----------
        options : dict
            Contains IO options.
        disc : dict
            Contains discretization parameters.
        bc : dict
            Contains boundary condition parameters.
        geometry : dict
            Contains geometry parameters.
        numerics : dict
            Contains numerics parameters.
        material : dict
            Contains material parameters.
        surface : dict
            Contains surface parameters.
        ic : dict
            Contains initial conditions.
        roughness : dict
            Contains roughness parameters.
        """

        self.options = options
        self.disc = disc
        self.bc = bc
        self.geometry = geometry
        self.numerics = numerics
        self.material = material
        self.surface = surface
        self.ic = ic
        self.roughness = roughness

        self.sanity_checks()

    def sanity_checks(self):

        self.check_options()
        self.check_disc()
        self.check_geometry()
        self.check_numerics()
        self.check_material()
        self.check_bc()
        if self.roughness is not None:
            self.check_roughness()
        if self.ic is not None:
            self.check_ic()
        if self.surface is not None:
            self.check_surface()

        print("Sanity checks completed. Start simulation!")
        print(60 * "-")

    def run(self, out_dir="data", out_name=None, plot=False):
        """
        Starts the simulation.

        Parameters
        ----------
        out_dir : str
            Output directory (default: data).
        out_name : str
            NetCDF output filename (default: None)
        plot : bool
            On-the-fly plotting flag (default: False).

        """

        # global write options
        writeInterval = self.options['writeInterval']

        if "maxT" in self.numerics.keys():
            maxT = self.numerics["maxT"]
        else:
            maxT = np.inf

        if "maxIt" in self.numerics.keys():
            maxIt = self.numerics["maxIt"]
        else:
            maxIt = np.inf

        if "tol" in self.numerics.keys():
            tol = self.numerics["tol"]
        else:
            tol = 0.

        # Initial conditions
        q_init, t_init = self.get_initial_conditions()

        # intialize solution field
        self.q = ConservedField(self.disc,
                                self.bc,
                                self.geometry,
                                self.material,
                                self.numerics,
                                self.surface,
                                self.roughness,
                                q_init=q_init,
                                t_init=t_init)

        rank = self.q.comm.Get_rank()

        # time stamp of simulation start time
        self.tStart = datetime.now()

        i = 0
        self._write_mode = 0

        # Header line for screen output
        if rank == 0:
            print(f"{'Step':10s}\t{'Timestep':12s}\t{'Time':12s}\t{'Epsilon':12s}", flush=True)
            self.write_to_stdout(i, mode=self._write_mode)

        if plot:
            # on-the-fly plotting
            self.plot(writeInterval)
        else:
            nc = self.init_netcdf(out_dir, out_name, rank)

            while self._write_mode == 0:

                # Perform time update
                self.q.update(i)

                # increase time step
                i += 1

                # catch signals and execute signal handler
                signal.signal(signal.SIGINT, self.receive_signal)
                signal.signal(signal.SIGTERM, self.receive_signal)
                signal.signal(signal.SIGHUP, self.receive_signal)
                signal.signal(signal.SIGUSR1, self.receive_signal)
                signal.signal(signal.SIGUSR2, self.receive_signal)

                # convergence
                if i > 1 and self.q.eps < tol:
                    self._write_mode = 1
                    break

                # maximum time reached
                if round(self.q.time, 15) >= maxT:
                    self._write_mode = 2
                    break

                # maximum number of iterations reached
                if i >= maxIt:
                    self._write_mode = 3
                    break

                # NaN detected
                if self.q.isnan > 0:
                    self._write_mode = 4
                    break

                if i % writeInterval == 0:
                    self.write_to_netcdf(i, nc, mode=self._write_mode)
                    if rank == 0:
                        self.write_to_stdout(i, mode=self._write_mode)

            self.write_to_netcdf(i, nc, mode=self._write_mode)
            if rank == 0:
                self.write_to_stdout(i, mode=self._write_mode)

    def get_initial_conditions(self):
        """
        Return the initial field given by last frame of restart file
        or as defined through inputs.

        Returns
        -------
        np.array
            Inital field of conserved variables.
        tuple
            Inital time and timestep
        """

        if self.ic is None:
            # No ICs specified, fill with (rho0, 0, 0)
            q_init = np.zeros((3, self.disc["Nx"], self.disc["Ny"]))
            q_init[0] += self.material["rho0"]
            t_init = (0., self.numerics["dt"])
        elif self.ic["type"] == "restart":
            # ICs read from last frame of restart file
            q_init, t_init = self.read_last_frame()
        else:
            # ICs as specified in config file
            t_init = (0., self.numerics["dt"])
            q_init = np.zeros((3, self.disc["Nx"], self.disc["Ny"]))
            q_init[0] = self.material["rho0"]

            if self.ic["type"] == "perturbation":
                q_init[0, self.disc["Nx"] // 2, self.disc["Ny"] // 2] *= self.ic["factor"]
            elif self.ic["type"] == "longitudinal_wave":
                x = np.linspace(0 + self.disc["dx"] / 2, self.disc["Lx"] - self.disc["dx"] / 2, self.disc["Nx"])
                y = np.linspace(0 + self.disc["dy"] / 2, self.disc["Ly"] - self.disc["dy"] / 2, self.disc["Ny"])
                xx, yy = np.meshgrid(x, y, indexing="ij")
                k = 2. * np.pi / self.disc["Lx"] * self.ic["nwave"]
                q_init[1] += self.ic["amp"] * np.sin(k * xx)
            elif self.ic["type"] == "shear_wave":
                x = np.linspace(0 + self.disc["dx"] / 2, self.disc["Lx"] - self.disc["dx"] / 2, self.disc["Nx"])
                y = np.linspace(0 + self.disc["dy"] / 2, self.disc["Ly"] - self.disc["dy"] / 2, self.disc["Ny"])
                xx, yy = np.meshgrid(x, y, indexing="ij")
                k = 2. * np.pi / self.disc["Lx"] * self.ic["nwave"]
                q_init[2] += self.ic["amp"] * np.sin(k * xx)
            elif self.ic["type"] == "random":
                q_init[0] += np.random.normal(0., self.ic["stdDens"], size=(self.disc["Nx"], self.disc["Ny"]))
                q_init[1] = np.random.normal(0., self.ic["stdFlux"], size=(self.disc["Nx"], self.disc["Ny"]))
                q_init[2] = np.random.normal(0., self.ic["stdFlux"], size=(self.disc["Nx"], self.disc["Ny"]))
            elif self.ic["type"] == "gauss1D":
                x = np.linspace(0 + self.disc["dx"] / 2, self.disc["Lx"] - self.disc["dx"] / 2, self.disc["Nx"])
                y = np.linspace(0 + self.disc["dy"] / 2, self.disc["Ly"] - self.disc["dy"] / 2, self.disc["Ny"])
                xx, yy = np.meshgrid(x, y, indexing="ij")

                # 1D centered Gaussian
                mean = self.ic['mean']
                stdev = self.ic['stdev']
                lc = self.ic['cutoff']

                # region inside cutoff
                inner = np.logical_and(xx > mean - lc, xx < mean + lc)

                p1D = np.exp(-(xx[inner] - mean)**2 / (2 * stdev**2)) / np.sqrt(2 * np.pi * stdev**2)
                p1D -= np.amin(p1D)  # shift
                p1D /= np.amax(p1D)  # rescale

                q_init[0, inner] += p1D

        return q_init, t_init

    def read_last_frame(self):
        """
        Read last frame from restart file and use as initial values for new run.

        Returns
        -------
        np.array
            Solution field at last frame, used as inital field.
        tuple
            Total time and timestep of last frame.
        """

        file = Dataset(self.ic["file"], "r")

        rho = np.array(file.variables['rho'])[-1]
        jx = np.array(file.variables['jx'])[-1]
        jy = np.array(file.variables['jy'])[-1]
        dt = float(file.variables["dt"][-1])
        t = float(file.variables["time"][-1])

        q0 = np.zeros([3] + list(rho.shape))

        q0[0] = rho
        q0[1] = jx
        q0[2] = jy

        return q0, (t, dt)

    def init_netcdf(self, out_dir, out_name, rank):
        """
        Initialize netCDF4 file, create dimensions, variables and metadata.

        Parameters
        ----------
        out_dir : str
            Output directoy.
        out_name : str
            Filename prefix.
        rank : int
            Rank of the MPI communicator
        Returns
        -------
        netCDF4.Dataset
            Initialized dataset.

        """

        if rank == 0:
            if not(os.path.exists(out_dir)):
                os.makedirs(out_dir)

        if self.ic is None or self.ic["type"] != "restart":
            if rank == 0:
                if out_name is None:
                    # default unique filename with timestamp
                    timestamp = datetime.now().replace(microsecond=0).strftime("%Y-%m-%d_%H%M%S")
                    name = self.options["name"]
                    outfile = f"{timestamp}_{name}.nc"
                else:
                    # custom filename with zero padded number
                    tag = str(len([1 for f in os.listdir(out_dir) if f.startswith(out_name)]) + 1).zfill(4)
                    outfile = f"{out_name}-{tag}.nc"

                self.outpath = os.path.join(out_dir, outfile)
            else:
                self.outpath = None

            self.outpath = self.q.comm.bcast(self.outpath, root=0)

            # initialize NetCDF file
            parallel = False
            if self.q.comm.Get_size() > 1:
                parallel = True

            nc = Dataset(self.outpath, 'w', parallel=parallel, format='NETCDF3_64BIT_OFFSET')
            nc.restarts = 0
            nc.createDimension('x', self.disc["Nx"])
            nc.createDimension('y', self.disc["Ny"])
            nc.createDimension('step', None)

            # create unknown variable buffer as timeseries of 2D fields
            var0 = nc.createVariable('rho', 'f8', ('step', 'x', 'y'))
            var1 = nc.createVariable('jx', 'f8', ('step', 'x', 'y'))
            var2 = nc.createVariable('jy', 'f8', ('step', 'x', 'y'))
            var0.set_collective(True)
            var1.set_collective(True)
            var2.set_collective(True)

            # create scalar variables
            nc.createVariable('time', 'f8', ('step'))
            nc.createVariable('mass', 'f8', ('step'))
            nc.createVariable('vmax', 'f8', ('step'))
            nc.createVariable('vSound', 'f8', ('step'))
            nc.createVariable('dt', 'f8', ('step'))
            nc.createVariable('eps', 'f8', ('step'))
            nc.createVariable('ekin', 'f8', ('step'))

            # write metadata
            nc.setncattr(f"tStart-{nc.restarts}", self.tStart.strftime("%d/%m/%Y %H:%M:%S"))
            nc.setncattr("version", get_distribution('hans').version)

            disc = self.disc.copy()
            bc = self.bc.copy()

            categories = {"options": self.options,
                          "disc": disc,
                          "bc": bc,
                          "geometry": self.geometry,
                          "numerics": self.numerics,
                          "material": self.material}

            if self.surface is not None:
                categories["surface"] = self.surface

            if self.ic is not None:
                categories["ic"] = self.ic

            if self.roughness is not None:
                roughness_noNone = {k: (v if v is not None else "None") for k, v in self.roughness.items()}
                categories["roughness"] = roughness_noNone

            # reset modified input dictionaries
            bc["x0"] = "".join(bc["x0"])
            bc["x1"] = "".join(bc["x1"])
            bc["y0"] = "".join(bc["y0"])
            bc["y1"] = "".join(bc["y1"])

            del disc["nghost"]
            del disc["pX"]
            del disc["pY"]

            for name, cat in categories.items():
                for key, value in cat.items():
                    nc.setncattr(f"{name}_{key}", value)

            self.write_to_netcdf(0, nc, mode=0)

        else:
            # append to existing netCDF file
            parallel = False
            if self.q.comm.Get_size() > 1:
                parallel = True
            nc = Dataset(self.ic["file"], 'a', parallel=parallel, format='NETCDF3_64BIT_OFFSET')
            self.outpath = os.path.relpath(self.ic["file"])

            backup_file = f"{os.path.splitext(self.ic['file'])[0]}-{nc.restarts}.nc"

            # create backup
            if rank == 0:
                shutil.copy(self.ic["file"], backup_file)

            # increase restart counter
            nc.restarts += 1

            # append modified attributes
            nc.setncattr(f"tStart-{nc.restarts}", self.tStart.strftime("%d/%m/%Y %H:%M:%S"))
            for key, value in self.numerics.items():
                name = f"numerics_{key}-{nc.restarts}"
                nc.setncattr(name, value)

            nc.setncattr(f"ic_type-{nc.restarts}", "restart")
            nc.setncattr(f"ic_file-{nc.restarts}", backup_file)

        return nc

    def write_to_stdout(self, i, mode):
        """
        Write information about the current time step to stdout.

        Parameters
        ----------
        i : int
            Current time step.
        mode : int
            Writing mode (0: normal, 1: converged, 2: max time, 3: execution stopped).

        """
        print(f"{i:10d}\t{self.q.dt:.6e}\t{self.q.time:.6e}\t{self.q.eps:.6e}", flush=True)

        if mode == 1:
            print(f"\nSolution has converged after {i:d} steps.", flush=True)
        elif mode == 2:
            print(f"\nNo convergence within {i: d} steps. Stopping criterion: \
maximum time {self.numerics['maxT']: .1e} s reached.", flush=True)
        elif mode == 3:
            print(f"\nNo convergence within {i: d} steps. Stopping criterion: \
maximum number of iterations reached.", flush=True)
        elif mode == 4:
            print("Nan detetcted in solution. Execution stopped.", flush=True)
        elif mode == 5:
            print("Execution stopped.", flush=True)

        if mode > 0:
            walltime = datetime.now() - self.tStart
            print(f"Output written to: {self.outpath}", flush=True)
            print(f"Total wall clock time: {str(walltime).split('.')[0]}", end=" ", flush=True)
            print(f"(Performance: {i/walltime.total_seconds(): .2f} steps/s", end=" ", flush=True)
            print(f"on {self.q.comm.dims[0]} x {self.q.comm.dims[1]} MPI grid)", flush=True)

    def write_to_netcdf(self, i, nc, mode):
        """
        Append current solution field to netCDF file.

        Parameters
        ----------
        i : int
            Current time step.
        nc : netCDF4.Dataset
            NetCDF Dataset object.
        mode : int
            Writing mode (0: normal, 1: converged, 2: max time, 3: execution stopped).

        """

        step = nc.variables["rho"].shape[0]
        xrange, yrange = self.q.without_ghost

        nc.variables['rho'][step, xrange, yrange] = self.q.inner[0]
        nc.variables['jx'][step, xrange, yrange] = self.q.inner[1]
        nc.variables['jy'][step, xrange, yrange] = self.q.inner[2]

        nc.variables["time"][step] = self.q.time
        nc.variables["mass"][step] = self.q.mass
        nc.variables["vmax"][step] = self.q.vmax
        nc.variables["vSound"][step] = self.q.vSound
        nc.variables["dt"][step] = self.q.dt
        nc.variables["eps"][step] = self.q.eps
        nc.variables["ekin"][step] = self.q.ekin

        if mode > 0:
            nc.setncattr(f"tEnd-{nc.restarts}", datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
            nc.close()

    def receive_signal(self, signum, frame):
        """
        Signal handler. Catches signals send to the process and sets write mode to 3 (abort).

        Parameters
        ----------
        signum :
            signal code
        frame :
            Description of parameter `frame`.

        """

        if signum in [signal.SIGINT, signal.SIGTERM, signal.SIGHUP, signal.SIGUSR1]:
            self._write_mode = 5

    def plot(self, writeInterval):
        """
        Initialize on-the-fly plotting.

        Parameters
        ----------
        writeInterval : int
            Write interval for stdout in plotting mode.

        """

        fig, ax = plt.subplots(2, 2, figsize=(14, 9), sharex=True)

        Nx = self.disc["Nx"]
        dx = self.disc["dx"]
        x = np.arange(Nx) * dx + dx / 2

        ax[0, 0].plot(x, self.q.centerline_x[1])
        ax[0, 1].plot(x, self.q.centerline_x[2])
        ax[1, 0].plot(x, self.q.centerline_x[0])
        ax[1, 1].plot(x, Material(self.material).eos_pressure(self.q.centerline_x[0]))

        ax[0, 0].set_title(r'$j_x$')
        ax[0, 1].set_title(r'$j_y$')
        ax[1, 0].set_title(r'$\rho$')
        ax[1, 1].set_title(r'$p$')

        ax[1, 0].set_xlabel('distance x (m)')
        ax[1, 1].set_xlabel('distance x (m)')

        def init():
            pass

        _ = animation.FuncAnimation(fig,
                                    self.animate1D,
                                    100000,
                                    fargs=(fig, ax, writeInterval),
                                    interval=1,
                                    init_func=init,
                                    repeat=False)

        plt.show()

    def animate1D(self, i, fig, ax, writeInterval):
        """
        Animator function. Update solution and plots.

        Parameters
        ----------
        i : type
            Current time step.
        fig : matplotlib.figure
            Figure object.
        ax :  np.array
            Array containing the axes of the figure.
        writeInterval : int
            Write interval for stdout in plotting mode.

        """

        self.q.update(i)
        fig.suptitle('time = {:.2f} ns'.format(self.q.time * 1e9))

        ax[0, 0].lines[0].set_ydata(self.q.centerline_x[1])
        ax[0, 1].lines[0].set_ydata(self.q.centerline_x[2])
        ax[1, 0].lines[0].set_ydata(self.q.centerline_x[0])
        ax[1, 1].lines[0].set_ydata(Material(self.material).eos_pressure(self.q.centerline_x[0]))

        ax = adaptiveLimits(ax)

        if i % writeInterval == 0 and i > 0:
            print(f"{i:10d}\t{self.q.dt:.6e}\t{self.q.time:.6e}\t{self.q.eps:.6e}", flush=True)

    def check_options(self):
        """
        Sanity check for I/O options input.
        """
        print("Checking I/O options... ")

        try:
            writeInterval = int(self.options["writeInterval"])
            assert writeInterval > 0
        except KeyError:
            print("***Output interval not given, fallback to 1000")
            self.options["writeInterval"] = 1000
        except AssertionError:
            try:
                assert writeInterval != 0
            except AssertionError:
                print("***Output interval is zero. fallback to 1000")
                self.options["writeInterval"] = 1000
            else:
                print("***Output interval is negative. Converting to positive value.")
                writeInterval *= -1
                self.options["writeInterval"] = writeInterval

    def check_disc(self):
        """
        Sanity check for discretization input.
        [Nx, Ny] are required, then look for [Lx, Ly] or [dx, dy] (in that order).
        """
        print("Checking discretization... ")

        try:
            self.disc["Nx"] = int(self.disc['Nx'])
            assert self.disc["Nx"] > 0
        except KeyError:
            print("***Number of grid cells Nx not specified. Abort.")
            abort()
        except AssertionError:
            print("***Number of grid cells Nx must be larger than zero. Abort")
            abort()

        try:
            self.disc["Ny"] = int(self.disc['Ny'])
            assert self.disc["Ny"] > 0
        except KeyError:
            print("***Number of grid cells 'Ny' not specified. Abort.")
            abort()
        except AssertionError:
            print("***Number of grid cells 'Ny' must be larger than zero. Abort")
            abort()

        try:
            self.disc["Lx"] = float(self.disc["Lx"])
        except KeyError:
            try:
                self.disc["dx"] = float(self.disc["dx"])
            except KeyError:
                print("At least two of 'Nx' 'Lx', 'dx' must be given. Abort.")
                abort()
            else:
                self.disc["Lx"] = self.disc["dx"] * self.disc["Nx"]
        else:
            self.disc["dx"] = self.disc["Lx"] / self.disc["Nx"]

        try:
            self.disc["Ly"] = float(self.disc["Ly"])
        except KeyError:
            try:
                self.disc["dy"] = float(self.disc["dy"])
            except KeyError:
                print("At least two of 'Ny' 'Ly', 'dy' must be given. Abort.")
                abort()
            else:
                self.disc["Ly"] = self.disc["dy"] * self.disc["Ny"]
        else:
            self.disc["dy"] = self.disc["Ly"] / self.disc["Ny"]

    def check_geometry(self):
        """
        Sanity check for geometry input.
        """
        print("Checking geometry... ")

        if self.geometry["type"] in ["journal", "journal_x", "journal_y"]:
            self.geometry["CR"] = float(self.geometry["CR"])
            self.geometry["eps"] = float(self.geometry["eps"])
        elif self.geometry["type"] == "parabolic":
            self.geometry["hmin"] = float(self.geometry['hmin'])
            self.geometry["hmax"] = float(self.geometry['hmax'])
        elif self.geometry["type"] == "twin_parabolic":
            self.geometry["hmin"] = float(self.geometry['hmin'])
            self.geometry["hmax"] = float(self.geometry['hmax'])
        elif self.geometry["type"] in ["inclined", "inclined_x", "inclined_y"]:
            self.geometry["h1"] = float(self.geometry['h1'])
            self.geometry["h2"] = float(self.geometry['h2'])
        elif self.geometry["type"] == "inclined_pocket":
            self.geometry["h1"] = float(self.geometry['h1'])
            self.geometry["h2"] = float(self.geometry['h2'])
            self.geometry["hp"] = float(self.geometry['hp'])
            self.geometry["c"] = float(self.geometry['c'])
            self.geometry["l"] = float(self.geometry['l'])
            self.geometry["w"] = float(self.geometry['w'])

            try:
                self.geometry["t"] = float(self.geometry["t"])
            except KeyError:
                self.geometry["t"] = 0.

            try:
                assert 2 * self.geometry["t"] + self.geometry["l"] + self.geometry["c"] <= self.disc["Lx"]
                assert self.geometry["w"] <= self.disc["Ly"]
            except AssertionError:
                print("Size of the pocket is larger than the domain. Abort.")
                abort()

        elif self.geometry["type"] in ["half_sine", "half_sine_squared"]:
            self.geometry["h0"] = float(self.geometry['h0'])
            self.geometry["amp"] = float(self.geometry['amp'])
            self.geometry["num"] = float(self.geometry['num'])

        else:
            print(f"'{self.geometry['type']}' geometry not implemented")
            abort()

    def check_numerics(self):
        """
        Sanity check for numerics options.
        """
        print("Checking numerics options... ")

        try:
            self.numerics["integrator"] = self.numerics["integrator"]
            assert self.numerics["integrator"] in ["MC", "MC_bf", "MC_fb", "MC_alt", "LW", "RK3"]
        except KeyError:
            print("***Integrator not specified. Use default (MacCormack).")
            self.numerics["integrator"] = "MC"
        except AssertionError:
            print(f'***Unknown integrator \'{self.numerics["integrator"]}\'. Abort.')
            abort()

        if self.numerics["integrator"].startswith("MC"):
            try:
                self.numerics["fluxLim"] = float(self.numerics["fluxLim"])
            except KeyError:
                pass

        try:
            self.numerics["stokes"] = int(self.numerics["stokes"])
        except KeyError:
            print("***Boolean parameter 'stokes' not given. Use default (True).")
            self.numerics["stokes"] = 1

        try:
            self.numerics["adaptive"] = int(self.numerics["adaptive"])
        except KeyError:
            print("***Boolean parameter 'adaptive' not given. Use default (False).")
            self.numerics["adaptive"] = 0

        if self.numerics["adaptive"] == 1:
            try:
                self.numerics["C"] = float(self.numerics["C"])
            except KeyError:
                print("***CFL number not given. Use default (0.5).")
                self.numerics["C"] = 0.5

        try:
            self.numerics["dt"] = float(self.numerics["dt"])
        except KeyError:
            print("***Timestep not given. Use default (1e-10).")
            self.numerics["dt"] = 1e-10

        stopping_criteria = 0

        try:
            self.numerics["tol"] = float(self.numerics["tol"])
            stopping_criteria += 1
        except KeyError:
            pass

        try:
            self.numerics["maxT"] = float(self.numerics["maxT"])
            stopping_criteria += 1
        except KeyError:
            pass

        try:
            self.numerics["maxIt"] = int(self.numerics["maxIt"])
            stopping_criteria += 1
        except KeyError:
            pass

        if stopping_criteria == 0:
            print("***No stopping criterion given. Abort.")
            abort()

        if self.numerics["integrator"] == "RK3":
            self.disc["nghost"] = 2
        else:
            self.disc["nghost"] = 1

    def check_material(self):
        """
        Sanity check on material settings.
        """
        print("Checking material options... ")

        if self.material["EOS"] == "DH":
            self.material["rho0"] = float(self.material["rho0"])
            self.material["P0"] = float(self.material["P0"])
            self.material["C1"] = float(self.material["C1"])
            self.material["C2"] = float(self.material["C2"])
        elif self.material["EOS"] == "PL":
            self.material["rho0"] = float(self.material["rho0"])
            self.material["P0"] = float(self.material["P0"])
            self.material["alpha"] = float(self.material['alpha'])
        elif self.material["EOS"] == "vdW":
            self.material["M"] = float(self.material['M'])
            self.material["T"] = float(self.material['T0'])
            self.material["a"] = float(self.material['a'])
            self.material["b"] = float(self.material['b'])
        elif self.material["EOS"] == "Tait":
            self.material["rho0"] = float(self.material["rho0"])
            self.material["P0"] = float(self.material["P0"])
            self.material["K"] = float(self.material['K'])
            self.material["n"] = float(self.material['n'])
        elif self.material["EOS"] == "cubic":
            self.material["a"] = float(self.material['a'])
            self.material["b"] = float(self.material['b'])
            self.material["c"] = float(self.material['c'])
            self.material["d"] = float(self.material['d'])
        elif self.material["EOS"].startswith("Bayada"):
            self.material["cl"] = float(self.material["cl"])
            self.material["cv"] = float(self.material["cv"])
            self.material["rhol"] = float(self.material["rhol"])
            self.material["rhov"] = float(self.material["rhov"])
            self.material["shear"] = float(self.material["shear"])
            self.material["shearv"] = float(self.material["shearv"])
            self.material["rhov"] = float(self.material["rhov"])

        self.material["shear"] = float(self.material["shear"])
        self.material["bulk"] = float(self.material["bulk"])

        if "Pcav" in self.material.keys():
            self.material["Pcav"] = float(self.material["Pcav"])

        if "piezo" in self.material.keys():
            if self.material["piezo"] == "Barus":
                self.material["aB"] = float(self.material["aB"])
            elif self.material["piezo"] == "Vogel":
                self.material["rho0"] = float(self.material['rho0'])
                self.material["g"] = float(self.material["g"])
                self.material["mu_inf"] = float(self.material["mu_inf"])
                self.material["phi_inf"] = float(self.material["phi_inf"])
                self.material["BF"] = float(self.material["BF"])

        if "thinning" in self.material.keys():
            if self.material["thinning"] == "Eyring":
                self.material["tau0"] = float(self.material["tau0"])
            elif self.material["thinning"] == "Carreau":
                self.material["relax"] = float(self.material["relax"])
                self.material["a"] = float(self.material["a"])
                self.material["N"] = float(self.material["N"])
            elif self.material["thinning"] == "PL":
                self.material["shear"] = float(self.material["shear"])
                self.material["n"] = float(self.material["n"])

        if "PLindex" in self.material.keys():
            self.material["PLindex"] = float(self.material["PLindex"])
            if "PLmethod" in self.material.keys():
                self.material["PLmethod"] = self.material["PLmethod"]
            else:
                self.material["PLmethod"] = "exact"

    def check_surface(self):
        """
        Sanity check for surface input.
        """
        print("Checking surface parameters... ")

        if "lslip" in self.surface.keys():
            self.surface["lslip"] = float(self.surface["lslip"])
        else:
            self.surface["lslip"] = 0.

        if self.surface["type"] in ["stripes", "stripes_x", "stripes_y"]:
            try:
                self.surface["num"] = int(self.surface["num"])
            except KeyError:
                self.surface["num"] = 1
            try:
                self.surface["sign"] = int(self.surface["sign"])
            except KeyError:
                self.surface["sign"] = -1

    def check_ic(self):
        """
        Sanity check for initial conditions input.
        """
        print("Checking initial conditions... ")

        if self.ic["type"] != "restart":
            if self.ic["type"] == "perturbation":
                self.ic["factor"] = float(self.ic["factor"])
            elif self.ic["type"] in ["longitudinal_wave", "shear_wave"]:
                self.ic["amp"] = float(self.ic["amp"])
                if "nwave" in self.ic.keys():
                    self.ic["nwave"] = int(self.ic["nwave"])
                else:
                    self.ic["nwave"] = 1
            elif self.ic["type"] == "random":
                try:
                    self.ic["stdDens"] = float(self.ic["stdDens"])
                except KeyError:
                    self.ic["stdDens"] = 0.0
                try:
                    self.ic["stdFlux"] = float(self.ic["stdFlux"])
                except KeyError:
                    self.ic["stdFlux"] = 0.0
            elif self.ic["type"] == "gauss1D":
                try:
                    self.ic["mean"] = float(self.ic["mean"])
                except KeyError:
                    self.ic["mean"] = float(self.disc["Lx"]) / 2
                try:
                    self.ic["stdev"] = float(self.ic["stdev"])
                except KeyError:
                    self.ic["stdev"] = float(self.disc["Lx"]) / 40
                try:
                    self.ic["cutoff"] = float(self.ic["cutoff"])
                except KeyError:
                    self.ic["cutoff"] = 12 * self.ic['stdev']

    def check_roughness(self):
        """
        Sanity check for random roughness input.
        """
        print("Checking roughness parameters... ")

        if "file" not in self.roughness.keys():
            try:
                self.roughness["seed"] = int(self.roughness["seed"])
            except KeyError:
                self.roughness["seed"] = None
            try:
                self.roughness["Hurst"] = float(self.roughness["Hurst"])
            except KeyError:
                "Hurst exponent not given. Use default (0.8)."
                self.roughness["Hurst"] = 0.8
            try:
                self.roughness["rolloff"] = float(self.roughness["rolloff"])
            except KeyError:
                "Rollof not given. Use default (1.0)"
                self.roughness["rolloff"] = 1.0
            try:
                self.roughness["rmsHeight"] = float(self.roughness["rmsHeight"])
            except KeyError:
                self.roughness["rmsHeight"] = None
            try:
                self.roughness["rmsSlope"] = float(self.roughness["rmsSlope"])
            except KeyError:
                self.roughness["rmsSlope"] = None
            try:
                self.roughness["shortCutoff"] = float(self.roughness["shortCutoff"])
            except KeyError:
                self.roughness["shortCutoff"] = None
            try:
                self.roughness["longCutoff"] = float(self.roughness["longCutoff"])
            except KeyError:
                self.roughness["longCutoff"] = None

            if self.roughness["rmsHeight"] is None and self.roughness["rmsSlope"] is None:
                print('Neither rms height nor rms slope is defined! Abort.')
                abort()

    def check_bc(self):
        """
        Sanity check for boundary condition input.

        Parameters
        ----------
        bc : dict
            Boundary condition parameters read from yaml input file.
        disc : dict
            Discretization parameters.
        material : dict
            Material parameters.

        Returns
        -------
        dict
            Boundary condition parameters.

        """
        print("Checking boundary conditions... ")

        self.bc["x0"] = np.array(list(self.bc["x0"]))
        self.bc["x1"] = np.array(list(self.bc["x1"]))
        self.bc["y0"] = np.array(list(self.bc["y0"]))
        self.bc["y1"] = np.array(list(self.bc["y1"]))

        assert len(self.bc["x0"]) == 3
        assert len(self.bc["x1"]) == 3
        assert len(self.bc["y0"]) == 3
        assert len(self.bc["y1"]) == 3

        if "P" in self.bc["x0"] and "P" in self.bc["x1"]:
            self.disc["pX"] = 1
        else:
            self.disc["pX"] = 0

        if "P" in self.bc["y0"] and "P" in self.bc["y1"]:
            self.disc["pY"] = 1
        else:
            self.disc["pY"] = 0

        if "D" in self.bc["x0"]:
            if "px0" in self.bc.keys():
                px0 = float(self.bc["px0"])
                self.bc["rhox0"] = Material(self.material).eos_density(px0)
            else:
                self.bc["rhox0"] = self.material["rho0"]

        if "D" in self.bc["x1"]:
            if "px1" in self.bc.keys():
                px1 = float(self.bc["px1"])
                self.bc["rhox1"] = Material(self.material).eos_density(px1)
            else:
                self.bc["rhox1"] = self.material["rho0"]

        if "D" in self.bc["y0"]:
            if "py0" in self.bc.keys():
                py0 = float(self.bc["py0"])
                self.bc["rhoy0"] = Material(self.material).eos_density(py0)
            else:
                self.bc["rhoy0"] = self.material["rho0"]

        if "D" in self.bc["y1"]:
            if "py1" in self.bc.keys():
                py1 = float(self.bc["py1"])
                self.bc["rhoy1"] = Material(self.material).eos_density(py1)
            else:
                self.bc["rhoy1"] = self.material["rho0"]

        assert np.all((self.bc["x0"] == "P") == (self.bc["x1"] == "P")), "Inconsistent boundary conditions (x)"
        assert np.all((self.bc["y0"] == "P") == (self.bc["y1"] == "P")), "Inconsistent boundary conditions (y)"



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import numpy as np
import pytest

from hans.input import Input
from hans.plottools import DatasetSelector


@pytest.fixture(scope="session")
def setup(tmpdir_factory):
    config_file = os.path.join("tests", "inclined-slider1D_x_ideal-gas.yaml")
    tmp_dir = tmpdir_factory.mktemp("tmp")

    myTestProblem = Input(config_file).getProblem()
    myTestProblem.run(out_dir=tmp_dir)

    file = DatasetSelector("", mode="name", fname=[str(tmp_dir.join(os.path.basename(myTestProblem.outpath)))])

    data = file.get_centerline()

    yield data


def test_pressure(setup):

    p_ref = np.loadtxt(os.path.join("tests", "inclined-slider1D_ideal-gas_U50_s5.6e-4.dat"), unpack=True, usecols=(2,))

    for xdata, ydata in setup:
        p = ydata["p"]
        np.testing.assert_almost_equal(p / 1e6, p_ref / 1e6, decimal=1)


def test_density(setup):

    rho_ref = np.loadtxt(os.path.join("tests", "inclined-slider1D_ideal-gas_U50_s5.6e-4.dat"), unpack=True, usecols=(1,))

    for xdata, ydata in setup:
        rho = ydata["rho"]
        np.testing.assert_almost_equal(rho, rho_ref, decimal=1)



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


from argparse import ArgumentParser
import matplotlib.pyplot as plt
import matplotlib.ticker as tk
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

from hans.plottools import DatasetSelector


def get_parser():
    parser = ArgumentParser()
    parser.add_argument('-p', dest="path", default="data", help="path (default: data)")
    parser.add_argument('-s', '--save', dest="save", action="store_true", default=False,
                        help="save height distribution as binary file in NumPy format (default: False)")

    return parser


if __name__ == "__main__":

    parser = get_parser()
    args = parser.parse_args()

    files = DatasetSelector(args.path, mode="single")
    fn, = files.get_filenames()
    zdata, = files.get_height()

    fig, ax = plt.subplots(1, figsize=(6.4, 4.8), tight_layout=True)
    print("Plotting ", fn)

    im = ax.imshow(zdata.T, extent=(0, 1, 0, 1))

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.1)

    fmt = tk.ScalarFormatter(useMathText=True)
    fmt.set_powerlimits((0, 0))

    cbar = plt.colorbar(im, cax=cax, format=fmt, orientation="vertical")
    cbar.set_label(r"Gap height  $h$")

    ax.set_xlabel(r"$x/L_x$")
    ax.set_ylabel(r"$y/L_y$")

    if args.save:
        ofn = fn.rstrip('.nc') + "_height.npy"
        np.save(ofn, zdata.T)

    plt.show()



#
# Copyright 2020, 2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import numpy as np
from scipy.optimize import fsolve

from hans.field import VectorField, TensorField
from hans.material import Material


class SymStressField2D(VectorField):

    def __init__(self, disc, geometry, material, surface=None):
        """This class contains the averaged viscous stress tensor components (xx, yy, xy).
        Derived from VectorField.

        Parameters
        ----------
        disc : dict
            Discretization parameters.
        geometry : dict
            Geometry parameters.
        material : dict
            Material parameters.
        surface : dict
            Surface parameters (the default is None).

        """

        super().__init__(disc)

        self.geometry = geometry
        self.material = material
        self.surface = surface

        try:
            self.n = self.material["PLindex"]
        except KeyError:
            self.n = 1

    def set(self, q, h, Ls):
        """Set method for Newtonian stress tensor components.

        Parameters
        ----------
        q : numpy.ndarray
            Field of conserved variables.
        h : numpy.ndarray
            Field of height and height gradients.
        Ls : numpy.ndarray
            Field of slip lengths.

        """

        U = self.geometry['U']
        V = self.geometry['V']
        eta = Material(self.material).viscosity(U, V, q[0], h[0])
        zeta = self.material['bulk']

        v1 = zeta + 4 / 3 * eta
        v2 = zeta - 2 / 3 * eta

        if self.surface is None or self.surface["type"] == "full":
            self.field[0] = (-3*((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls**2*(U*q[0] - 2*q[1]))*v1*h[1]
                             - 3*((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                  + 6*Ls**2*(V*q[0] - 2*q[2])) * h[2]*v2)/(3*h[0]*q[0]*(h[0] + 6*Ls)*(h[0] + 2*Ls))

            self.field[1] = (-3*((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls**2*(U*q[0] - 2*q[1]))*v2*h[1]
                             - 3*((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                  + 6*Ls**2*(V*q[0] - 2*q[2])) * h[2]*v1)/(3*h[0]*q[0]*(h[0] + 6*Ls)*(h[0] + 2*Ls))

            self.field[2] = -eta*(((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0] + 6*Ls ** 2*(V*q[0] - 2*q[2]))*h[1]
                                  + ((U*q[0] - 3*q[1])*h[0] ** 2 + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls ** 2*(U*q[0] - 2*q[1]))
                                  * h[2])/(h[0]*q[0]*(h[0] + 6*Ls)*(h[0] + 2*Ls))

        else:
            # all other types than 'full' consider slip only at the top surface
            self.field[0] = (-3*v1*((U*q[0] - 3*q[1])*h[0]**2
                                    + 3*Ls*(U*q[0] - 2*q[1])*h[0] + 6*Ls**2*(U*q[0] - q[1]))*h[1]
                             - 3*v2*h[2] * ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 2*q[2])*h[0]
                                            + 6*Ls**2*(V*q[0] - q[2])))/(3*h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))

            self.field[1] = (-3*v2*((U*q[0] - 3*q[1])*h[0]**2
                                    + 3*Ls*(U*q[0] - 2*q[1])*h[0] + 6*Ls**2*(U*q[0] - q[1]))*h[1]
                             - 3*v1*h[2] * ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 2*q[2])*h[0]
                                            + 6*Ls**2*(V*q[0] - q[2])))/(3*h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))

            self.field[2] = -eta*(((V*q[0] - 3*q[2])*h[0]**2
                                   + 3*Ls*(V*q[0] - 2*q[2])*h[0] + 6*Ls**2*(V*q[0] - q[2]))*h[1]
                                  + ((U*q[0] - 3*q[1]) * h[0]**2 + 3*Ls*(U*q[0] - 2*q[1])*h[0]
                                     + 6*Ls**2*(U*q[0] - q[1]))*h[2])/(h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))


class SymStressField3D(TensorField):

    def __init__(self, disc, geometry, material, surface=None):
        """This class contains all viscous stress tensor components
        evaluated at the top and bottom respectively. Derived from TensorField.

        Parameters
        ----------
        disc : dict
            Discretization parameters.
        geometry : dict
            Geometry parameters.
        material : dict
            Material parameters.
        surface : dict
            Surface parameters (the default is None).

        """

        super().__init__(disc)

        self.geometry = geometry
        self.material = material
        self.surface = surface

        try:
            self.n = self.material["PLindex"]
            self.PL = True
        except KeyError:
            self.PL = False

    def set(self, q, h, Ls, bound):
        """Wrapper function around set methods for wall stress tensor components.

        Parameters
        ----------
        q : numpy.ndarray
            Field of conserved variables.
        h : numpy.ndarray
            Field of height and height gradients.
        Ls : numpy.ndarray
            Field of slip lengths.
        bound : str
            Flag for bottom or top wall.

        """

        if self.PL:
            self.set_PowerLaw(q, h, bound)
        else:
            self.set_Newton(q, h, Ls, bound)

    def set_Newton(self, q, h, Ls, bound):
        """Set method for Newtonian stress tensor components.

        Parameters
        ----------
        q : numpy.ndarray
            Field of conserved variables.
        h : numpy.ndarray
            Field of height and height gradients.
        Ls : numpy.ndarray
            Field of slip lengths.
        bound : str
            Flag for bottom or top wall.

        """

        U = self.geometry['U']
        V = self.geometry['V']
        eta = Material(self.material).viscosity(U, V, q[0], h[0])
        zeta = self.material['bulk']

        v1 = zeta + 4 / 3 * eta
        v2 = zeta - 2 / 3 * eta

        if bound == "top":
            if self.surface is None or self.surface["type"] == "full":
                # all other types than 'full' consider slip only at the top surface
                self.field[0] = -2*(((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls**2*(U*q[0] - 2*q[1]))*v1*h[1]
                                    + ((V*q[0] - 3*q[2]) * h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                       + 6*Ls**2*(V*q[0] - 2*q[2]))*h[2]*v2) * (h[0] + Ls)/(h[0]*(h[0] + 2*Ls)**2*q[0]*(h[0] + 6*Ls))

                self.field[1] = -2*(h[0] + Ls)*(((U*q[0] - 3*q[1])*h[0]**2
                                                 + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls**2*(U*q[0] - 2*q[1]))*v2*h[1]
                                                + ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                                    + 6*Ls**2*(V*q[0] - 2*q[2]))*h[2]*v1) / (h[0]*(h[0] + 2*Ls)**2*q[0]*(h[0] + 6*Ls))

                self.field[2] = -2*(h[0] + Ls)*v2*(((U*q[0] - 3*q[1])*h[0]**2
                                                    + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls**2*(U*q[0] - 2*q[1]))*h[1]
                                                   + ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                                      + 6*Ls**2*(V*q[0] - 2*q[2]))*h[2])/(h[0]*(h[0] + 2*Ls)**2*q[0]*(h[0] + 6*Ls))

                self.field[3] = 2*(q[0]*V*h[0] - 3*h[0]*q[2] - 6*q[2]*Ls)*eta/(q[0]*(h[0]**2 + 8*Ls*h[0] + 12*Ls**2))

                self.field[4] = 2*(q[0]*U*h[0] - 3*h[0]*q[1] - 6*q[1]*Ls)*eta/(q[0]*(h[0]**2 + 8*Ls*h[0] + 12*Ls**2))

                self.field[5] = -2*eta*(h[0] + Ls)*(((V*q[0] - 3*q[2])*h[0]**2
                                                     + 3*Ls*(V*q[0] - 4*q[2])*h[0] + 6*Ls**2*(V*q[0] - 2*q[2]))*h[1]
                                                    + ((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 4*q[1])*h[0]
                                                       + 6*Ls**2*(U*q[0] - 2*q[1]))*h[2])/(h[0]*(h[0] + 2*Ls)**2*q[0]*(h[0] + 6*Ls))
            else:
                # all other types than 'full' consider slip only at the top surface
                self.field[0] = (-6*v1*((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 2*q[1])*h[0] + 6*Ls**2*(U*q[0] - q[1]))*h[1]
                                 - 6*v2*h[2] * ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 2*q[2])*h[0]
                                                + 6*Ls**2*(V*q[0] - q[2])))/(3*h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))

                self.field[1] = (-6*v2*((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 2*q[1])*h[0] + 6*Ls**2*(U*q[0] - q[1]))*h[1]
                                 - 6*v1*h[2] * ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 2*q[2])*h[0]
                                                + 6*Ls**2*(V*q[0] - q[2])))/(3*h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))

                self.field[2] = -2*v2*(((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 2*q[1])*h[0] + 6*Ls**2*(U*q[0] - q[1]))*h[1] +
                                       h[2]*((V*q[0] - 3*q[2])*h[0]**2
                                             + 3*Ls*(V*q[0] - 2*q[2])*h[0] + 6*Ls**2*(V*q[0] - q[2])))/(h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))

                self.field[3] = 2*(V*q[0] - 3*q[2])*eta/(q[0]*(4*Ls + h[0]))

                self.field[4] = 2*(U*q[0] - 3*q[1])*eta/(q[0]*(4*Ls + h[0]))

                self.field[5] = -2*eta*(((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 2*q[2])*h[0] + 6*Ls**2*(V*q[0] - q[2]))*h[1]
                                        + ((U*q[0] - 3*q[1])*h[0]**2 + 3*Ls*(U*q[0] - 2*q[1])*h[0]
                                           + 6*Ls**2*(U*q[0] - q[1]))*h[2])/(h[0]*(4*Ls + h[0])*q[0]*(Ls + h[0]))

        elif bound == "bottom":
            if self.surface is None or self.surface["type"] == "full":
                # surface type "full" means that both surface are slippery with equal slip length
                self.field[0] = -2*(((U*q[0] - 3*q[1])*h[0]**2
                                     + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls ** 2*(U*q[0] - 2*q[1]))*v1*h[1]
                                    + ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                       + 6*Ls**2*(V*q[0] - 2*q[2]))*h[2]*v2)*Ls/(h[0]*(h[0] + 2*Ls) ** 2*q[0]*(h[0] + 6*Ls))

                self.field[1] = -2*(((U*q[0] - 3*q[1])*h[0]**2
                                     + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls ** 2*(U*q[0] - 2*q[1]))*v2*h[1]
                                    + ((V*q[0] - 3*q[2])*h[0]**2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                       + 6*Ls**2*(V*q[0] - 2*q[2]))*h[2]*(zeta + (4*eta)/3))*Ls/(h[0]*(h[0] + 2*Ls) ** 2*q[0]*(h[0] + 6*Ls))

                self.field[2] = -2*v2*Ls*(((U*q[0] - 3*q[1])*h[0]**2
                                           + 3*Ls*(U*q[0] - 4*q[1])*h[0] + 6*Ls ** 2*(U*q[0] - 2*q[1]))*h[1]
                                          + ((V*q[0] - 3*q[2])*h[0] ** 2 + 3*Ls*(V*q[0] - 4*q[2])*h[0]
                                             + 6*Ls ** 2*(V*q[0] - 2*q[2]))*h[2])/(h[0]*(h[0] + 2*Ls) ** 2*q[0]*(h[0] + 6*Ls))

                self.field[3] = -2*(2*q[0]*V*h[0] + 6*q[0]*V*Ls - 3*h[0]*q[2] - 6*q[2]*Ls)*eta/(q[0]*(h[0] ** 2 + 8*Ls*h[0] + 12*Ls ** 2))

                self.field[4] = -2*(2*q[0]*U*h[0] + 6*q[0]*U*Ls - 3*h[0]*q[1] - 6*q[1]*Ls)*eta/(q[0]*(h[0] ** 2 + 8*Ls*h[0] + 12*Ls ** 2))

                self.field[5] = -2*eta*(((V*q[0] - 3*q[2])*h[0]**2
                                         + 3*Ls*(V*q[0] - 4*q[2])*h[0] + 6*Ls**2*(V*q[0] - 2*q[2]))*h[1]
                                        + ((U*q[0] - 3*q[1])*h[0] ** 2 + 3*Ls*(U*q[0] - 4*q[1])*h[0]
                                           + 6*Ls ** 2*(U*q[0] - 2*q[1]))*h[2])*Ls/(h[0]*(h[0] + 2*Ls) ** 2*q[0]*(h[0] + 6*Ls))
            else:
                # all other types than 'full' consider slip only at the top surface
                self.field[3] = -2*(6*q[0]*Ls*V + 2*q[0]*V*h[0] - 6*q[2]*Ls - 3*h[0]*q[2])*eta/(h[0]*q[0]*(4*Ls + h[0]))
                self.field[4] = -2*(6*q[0]*Ls*U + 2*q[0]*U*h[0] - 6*q[1]*Ls - 3*h[0]*q[1])*eta/(h[0]*q[0]*(4*Ls + h[0]))

    def set_PowerLaw(self, q, h, bound):
        """Set method for power law stress tensor components.

        Parameters
        ----------
        q : numpy.ndarray
            Field of conserved variables.
        h : numpy.ndarray
            Field of height and height gradients.
        bound : str
            Flag for bottom or top wall.

        """

        U = self.geometry['U']
        V = self.geometry['V']
        eta = self.material["shear"]

        n = self.n

        u_mean = q[1] / q[0]
        v_mean = q[2] / q[0]

        if self.material["PLmethod"] == "approx":
            zmaxu, zmaxv = self.approximate_zmax(U, V, u_mean, v_mean)
        else:
            zmaxu, zmaxv = self.solve_zmax(U, V, u_mean, v_mean)

        zmaxu *= h[0]
        zmaxv *= h[0]

        U_crit = u_mean * (1 + 2 * n) / (1 + n)
        V_crit = v_mean * (1 + 2 * n) / (1 + n)

        # case 1: U <= U_crit; V <= V_crit
        maskU1 = U <= U_crit
        maskV1 = V <= V_crit

        # case 2: U > U_crit; V > V_crit
        maskU2 = U > U_crit
        maskV2 = V > V_crit

        if bound == "top":
            # case 1
            s0_1 = -2 * ((1+2*n)*power(-zmaxu+h[0], 1/n)*(q[0]*U*zmaxu-q[1]*h[0])*h[1]) / (n*q[0]*(power(zmaxu, 2+1/n)
                                                                                                   + power(-zmaxu+h[0], 2+1/n)))

            s1_1 = -2 * ((1+2*n)*power(-zmaxv+h[0], 1/n)*(q[0]*V*zmaxv-q[2]*h[0])*h[2]) / (n*q[0]*(power(zmaxv, 2+1/n)
                                                                                                   + power(-zmaxv+h[0], 2+1/n)))

            s3_1 = ((1+2*n)*power(-zmaxv+h[0], 1/n)*(q[0]*V*zmaxv-q[2]*h[0]))/(n*q[0]*(power(zmaxv, 2+1/n)+power(-zmaxv+h[0], 2+1/n)))

            s4_1 = ((1+2*n)*power(-zmaxu+h[0], 1/n)*(q[0]*U*zmaxu-q[1]*h[0]))/(n*q[0]*(power(zmaxu, 2+1/n)+power(-zmaxu+h[0], 2+1/n)))

            s5_u1 = -s4_1 * h[2]
            s5_v1 = -s3_1 * h[1]

            # case 2
            s0_2 = -2*((1+2*n)*(q[1]-U*q[0])*(power(zmaxu, 1+1/n)-power(zmaxu-h[0], 1+1/n)) *
                       (power(zmaxu, 2+1/n)*n-power(zmaxu - h[0], 1+1/n)
                        * (zmaxu*n+(1+n)*h[0]))*h[1])/(q[0]*(n*power(zmaxu-h[0], 2+1/n)
                                                             + power(zmaxu, 1+1/n)*(-zmaxu*n+(1+2*n)*h[0]))**2)

            s1_2 = -2*((1+2*n)*(q[2]-V*q[0])*(power(zmaxv, 1+1/n)-power(zmaxv-h[0], 1+1/n)) *
                       (power(zmaxv, 2+1/n)*n-power(zmaxv - h[0], 1+1/n)
                        * (zmaxv*n+(1+n)*h[0]))*h[2])/(q[0]*(n*power(zmaxv-h[0], 2+1/n)
                                                             + power(zmaxv, 1+1/n)*(-zmaxv*n+(1+2*n)*h[0]))**2)

            s3_2 = ((1+1/n)*(1+2*n)*(q[2]-V*q[0])*power(zmaxv-h[0], 1/n)*h[0]) / \
                (q[0]*(n*power(zmaxv-h[0], 2+1/n)+power(zmaxv, 1+1/n)*(-zmaxv*n+(1+2*n)*h[0])))

            s4_2 = ((1+1/n)*(1+2*n)*(q[1]-U*q[0])*power(zmaxu-h[0], 1/n)*h[0]) / \
                (q[0]*(n*power(zmaxu-h[0], 2+1/n)+power(zmaxu, 1+1/n)*(-zmaxu*n+(1+2*n)*h[0])))

            s5_u2 = (1+2*n)/q[0]*(-(((1+2*n)*(q[1]-U*q[0])*(power(zmaxu, 1+1/n)
                                                            - power(zmaxu-h[0], 1+1/n))**2 * h[0]*h[2]) /
                                    (n*power(zmaxu-h[0], 2+1/n) + power(zmaxu, 1+1/n)*(-zmaxu * n+(1+2*n)*h[0]))**2)
                                  + ((q[1]-U*q[0])*(power(zmaxu, 1+1/n)-power(zmaxu-h[0], 1+1/n))*h[2]) /
                                  (n*power(zmaxu-h[0], 2+1/n)+power(zmaxu, 1+1/n)*(-zmaxu*n+(1+2*n)*h[0])))

            s5_v2 = (1+2*n)/q[0]*(-(((1+2*n)*(q[2]-V*q[0])*(power(zmaxv, 1+1/n)
                                                            - power(zmaxv-h[0], 1+1/n))**2 * h[0]*h[1]) /
                                    (n*power(zmaxv-h[0], 2+1/n) + power(zmaxv, 1+1/n)*(-zmaxv * n+(1+2*n)*h[0]))**2)
                                  + ((q[2]-V*q[0])*(power(zmaxv, 1+1/n)-power(zmaxv-h[0], 1+1/n))*h[1]) /
                                  (n*power(zmaxv-h[0], 2+1/n)+power(zmaxv, 1+1/n)*(-zmaxv*n+(1+2*n)*h[0])))

            self.field[0, maskU1] = eta * power(s0_1[maskU1], n)
            self.field[0, maskU2] = eta * power(s0_2[maskU2], n)

            self.field[1, maskV1] = eta * power(s1_1[maskV1], n)
            self.field[1, maskV2] = eta * power(s1_2[maskV2], n)

            self.field[3, maskV1] = eta * power(s3_1[maskV1], n)
            self.field[3, maskV2] = eta * power(s3_2[maskV2], n)

            self.field[4, maskU1] = eta * power(s4_1[maskU1], n)
            self.field[4, maskU2] = eta * power(s4_2[maskU2], n)

            self.field[5, maskU1] = eta * power(s5_u1[maskU1], n)
            self.field[5, maskU2] = eta * power(s5_u2[maskU2], n)

            self.field[5, maskV1] += eta * power(s5_v1[maskV1], n)
            self.field[5, maskV2] += eta * power(s5_v2[maskV2], n)

        elif bound == "bottom":
            # case 1:
            s3_1 = -(((1+2*n)*power(zmaxv, 1/n)*(q[0]*V*zmaxv-q[2]*h[0]))/(n*q[0]*(power(zmaxv, 2+1/n)+power(-zmaxv+h[0], 2+1/n))))

            s4_1 = -(((1+2*n)*power(zmaxu, 1/n)*(q[0]*U*zmaxu-q[1]*h[0]))/(n*q[0]*(power(zmaxu, 2+1/n)+power(-zmaxu+h[0], 2+1/n))))

            # case 2:
            s3_2 = (power(zmaxv, 1/n)*(1+1/n)*(1+2*n)*(q[2]-V*q[0])*h[0]) / (q[0]*(n*power(zmaxv-h[0], 2+1/n) +
                                                                                   power(zmaxv, 1+1/n)*(-zmaxv*n+(1+2*n)*h[0])))

            s4_2 = (power(zmaxu, 1/n)*(1+1/n)*(1+2*n)*(q[1]-U*q[0])*h[0]) / (q[0]*(n*power(zmaxu-h[0], 2+1/n) +
                                                                                   power(zmaxu, 1+1/n)*(-zmaxu*n+(1+2*n)*h[0])))

            self.field[3, maskV1] = eta * power(s3_1[maskV1], n)
            self.field[3, maskV2] = eta * power(s3_2[maskV2], n)

            self.field[4, maskU1] = eta * power(s4_1[maskU1], n)
            self.field[4, maskU2] = eta * power(s4_2[maskU2], n)

    def solve_zmax(self, U, V, um, vm):
        """Calculate the location of maximum velocity in z-direction to compute the power law stresses.

        Parameters
        ----------
        U : float
            Sliding velocity of the lower surface in x-direction.
        V : float
            Sliding velocity of the lower surface in y-direction.
        Um : numpy.ndarray
            Array of mean velocities in x-direction (jx/rho).
        Vm : numpy.ndarray
            Array of mean velocities in y-direction (jy/rho).

        Returns
        -------
        numpy.ndarray
            Nondimensional z-coordinate of velocity maximum in x-direction
        numpy.ndarray
            Nondimensional z-coordinate of velocity maximum in y-direction
        """

        Nx, Ny = vm.shape

        if V == 0:
            zmaxv = np.ones_like(vm) * 0.5
        else:
            vn = vm / V
            zmaxv = np.ones_like(vm)
            init = 0.5 - 1 / (12*self.n*(vn - 0.5))

            for i in range(Nx):
                for j in range(Ny):
                    if vn[i, j] >= (1 + self.n) / (1 + 2 * self.n):
                        zmaxv[i, j] = fsolve(zmax_nleq_case1, init[i, j], args=(self.n, vn[i, j]))[0]
                    else:
                        zmaxv[i, j] = fsolve(zmax_nleq_case2, init[i, j], args=(self.n, vn[i, j]))[0]

        if U == 0:
            zmaxu = np.ones_like(um) * 0.5
        else:
            un = um / U
            zmaxu = np.ones_like(um)
            init = 0.5 - 1 / (12*self.n*(un - 0.5))

            for i in range(Nx):
                for j in range(Ny):
                    if un[i, j] >= (1 + self.n) / (1 + 2 * self.n):
                        zmaxu[i, j] = fsolve(zmax_nleq_case1, init[i, j], args=(self.n, un[i, j]))[0]
                    else:
                        zmaxu[i, j] = fsolve(zmax_nleq_case2, init[i, j], args=(self.n, un[i, j]))[0]

        # upper and lower bound for asymptotic behavior
        zmaxu = np.minimum(zmaxu, 1e5)
        zmaxu = np.maximum(zmaxu, -1e5)
        zmaxv = np.minimum(zmaxv, 1e5)
        zmaxv = np.maximum(zmaxv, -1e5)

        return zmaxu, zmaxv

    def approximate_zmax(self, U, V, Um, Vm):
        """Approximate the location of maximum velocity in z-direction to compute the power law stresses.
        Exact values have been found numerically for both cases, but a good approximation
        is given by a hyperbola with vertical asymptote at jx/rho = U/2 (Couette profile).

        Parameters
        ----------
        U : float
            Sliding velocity of the lower surface in x-direction.
        V : float
            Sliding velocity of the lower surface in y-direction.
        Um : numpy.ndarray
            Array of mean velocities in x-direction (jx/rho).
        Vm : numpy.ndarray
            Array of mean velocities in y-direction (jy/rho).

        Returns
        -------
        numpy.ndarray
            Nondimensional z-coordinate of velocity maximum in x-direction
        numpy.ndarray
            Nondimensional z-coordinate of velocity maximum in y-direction
        """

        a = 0.5
        bu = -U / (12 * self.n)
        cu = U / 2

        bv = -V / (12 * self.n)
        cv = V / 2

        if V == 0:
            zmaxv = np.ones_like(Vm) * 0.5
        else:
            zmaxv = a + bv / (Vm - cv)

        if U == 0:
            zmaxu = np.ones_like(Um) * 0.5
        else:
            zmaxu = a + bu / (Um - cu)

        # upper and lower bound for asymptotic behavior
        zmaxu = np.minimum(zmaxu, 1e5)
        zmaxu = np.maximum(zmaxu, -1e5)
        zmaxv = np.minimum(zmaxv, 1e5)
        zmaxv = np.maximum(zmaxv, -1e5)

        return zmaxu, zmaxv


def zmax_nleq_case1(zmax, n, un):
    """Definition of nonlinear equation (Case 1) to be solved for zmax with scipy.optimize.fsolve.

    Parameters
    ----------
    zmax : float
        z-location of velocity maximum
    n : float
        Power-law exponent
    un : float
        non-dimensional height-averaged velocity.

    Returns
    -------
    float
        Function value f(zmax) = 0 for root-finding.
    """

    # Case1
    return power((1+n)/(n * (-power(zmax, 1+1/n)
                             + power(1. - zmax, 1+1/n))), n) - power(((1+2*n)*(un - zmax))/(n*(power(zmax, 2+1/n)
                                                                                               + power(1. - zmax, 2+1/n))), n)


def zmax_nleq_case2(zmax, n, un):
    """Definition of nonlinear equation (Case 2) to be solved for zmax with scipy.optimize.fsolve.

    Parameters
    ----------
    zmax : float
        z-location of velocity maximum
    n : float
        Power-law exponent
    un : float
        non-dimensional height-averaged velocity.

    Returns
    -------
    float
        Function value f(zmax) = 0 for root-finding.
    """

    return 1. - (un*(1+2*n)*(power(zmax, 1+1/n)-power(zmax-1, 1+1/n))) / (power(zmax, 2+1/n)*n - power(zmax-1, 1+1/n)*(zmax*n+(1+n)))


def power(base, exponent):

    complex_base = base.astype('complex128')
    result = np.float_power(complex_base, exponent)

    return result.real



#
#


import sys

file_lines = open(sys.argv[1], 'r').readlines()
header_lines = sys.stdin.readlines()

while file_lines[0].startswith('#'):
    file_lines = file_lines[1:]

file_lines.insert(0, '#\n')
for header_line in header_lines[::-1]:
    file_lines.insert(0, '# {}'.format(header_line).strip() + '\n')
file_lines.insert(0, '#\n')

open(sys.argv[1], 'w').writelines(file_lines)



#
# Copyright 2021-2022 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from argparse import ArgumentParser
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tk
import matplotlib.animation as animation
from mpl_toolkits.axes_grid1 import make_axes_locatable


from hans.plottools import DatasetSelector


def get_parser():
    parser = ArgumentParser()
    parser.add_argument('-p', dest="path", default="data", help="path (default: data)")
    parser.add_argument('-v', dest="key", default=None, choices=[None, "rho", "p", "jx", "jy"], help="variable (default: p)")

    return parser


def update_grids(i, A, t):
    """
    Updates the plot in animation

    Parameters
    ----------
    i : int
        iterator
    A : list
        contains 2D arrays of field variables
    t : list
        contains time
    """

    if type(A) is dict:
        for ax, k in zip(fig.axes, A.keys()):
            im, = ax.get_images()
            im.set_array(A[k][i].T)
            if i > 0:
                im.set_clim(vmin=np.amin(A[k][:i]), vmax=np.amax(A[k][:i]))
    else:
        ax = fig.axes[0]
        im, = ax.get_images()
        im.set_array(A[i].T)

        if i > 0:
            im.set_clim(vmin=np.amin(A[:i]), vmax=np.amax(A[:i]))

    fig.suptitle("Time: {:.1g} s".format(t[i]))


if __name__ == "__main__":

    ylabels = {"rho": r"Density $\rho$",
               "p": r"Pressure $p$",
               "jx": r"Momentum density $j_x$",
               "jy": r"Momentum denisty $j_y$"}

    parser = get_parser()
    args = parser.parse_args()

    files = DatasetSelector(args.path, mode="single")
    fn, = files.get_filenames()
    time, zdata = files.get_fields(key=args.key)[0]
    print("Animating ", fn)

    if args.key is None:
        fig, ax = plt.subplots(2, 2, figsize=(12.8, 9.6))
        for key, axis in zip(zdata.keys(), ax.flat):
            im = axis.imshow(zdata[key][0].T, extent=(0, 1, 0, 1))

            # colorbar
            divider = make_axes_locatable(axis)
            cax = divider.append_axes("right", size="5%", pad=0.3)
            fmt = tk.ScalarFormatter(useMathText=True)
            fmt.set_powerlimits((0, 0))
            cbar = plt.colorbar(im, cax=cax, format=fmt, orientation="vertical")
            cbar.set_label(ylabels[key])

            # set x-/y-labels
            axis.set_xlabel(r'$x/L_x$')
            axis.set_ylabel(r'$y/L_y$')
    else:
        fig, ax = plt.subplots(figsize=(6.4, 4.8))

        # initial plot
        im = ax.imshow(zdata[0].T, extent=(0, 1, 0, 1))

        # colorbar
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.3)
        fmt = tk.ScalarFormatter(useMathText=True)
        fmt.set_powerlimits((0, 0))
        cbar = plt.colorbar(im, cax=cax, format=fmt, orientation="vertical")
        cbar.set_label(ylabels[args.key])

        # set x-/y-labels
        ax.set_xlabel(r'$x/L_x$')
        ax.set_ylabel(r'$y/L_y$')

    ani = animation.FuncAnimation(fig, update_grids, frames=len(time), fargs=(zdata, time), interval=100, repeat=True)

    plt.show()



#
# Copyright 2020-2021 Hannes Holey
#
# ### MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import os
from mpi4py import MPI
from argparse import ArgumentParser

from hans.input import Input


def get_parser():
    """Parses input arguments from command line.

    Returns
    -------
    ArgumentParser
        Parser object.

    """

    parser = ArgumentParser()

    # required arguments
    required = parser.add_argument_group('required arguments')
    required.add_argument('-i', '--input',
                          dest="filename",
                          help="path to input file",
                          required=True)

    # optional arguments
    parser.add_argument('-p', '--plot',
                        dest='plot',
                        default=False,
                        help="on-the-fly plot option",
                        action='store_true')
    parser.add_argument('-r', '--restart',
                        dest="restart_file",
                        default=None,
                        help="restart simulation from last step of specified file")
    parser.add_argument('-o', '--output',
                        dest="out_dir",
                        default="data",
                        help="output directory (default: ./data)")
    parser.add_argument('-n', '--name',
                        dest="out_name",
                        default=None,
                        help="output filename (without extension *.nc)")

    return parser


if __name__ == "__main__":

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()

    if rank == 0:
        parser = get_parser()
        args = parser.parse_args()

        if args.plot and comm.Get_size() > 1:
            print("Live-plotting not implemented for parallel execution", flush=True)
            comm.Abort()

        inputFile = os.path.join(os.getcwd(), args.filename)
        if args.restart_file is not None:
            restartFile = os.path.join(os.getcwd(), args.restart_file)
        else:
            restartFile = None

        myProblem = Input(inputFile, restartFile).getProblem()

    else:
        myProblem = None
        args = None

    args = comm.bcast(args, root=0)
    myProblem = comm.bcast(myProblem, root=0)
    myProblem.run(out_dir=args.out_dir, out_name=args.out_name, plot=args.plot)
